#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include "MotorGlueComponent.hh"

using ::testing::_;
using ::testing::IsNull;
using ::testing::NotNull;
using ::testing::Throw;
using ::testing::Return;

namespace {

class MotorMock : public Hal::Motor {
public:
    MOCK_METHOD(void, Home, (), (override));
    MOCK_METHOD(bool, IsHoming, (), (override));
    MOCK_METHOD(void, Move, (double pos), (override));
    MOCK_METHOD(bool, IsMoving, (), (override));
    MOCK_METHOD(void, Stop, (), (override));

    MOCK_METHOD(void, HomingChanged, (std::function<void()> event), (override));
    MOCK_METHOD(void, MovingChanged, (std::function<void()> event), (override));

    auto HomingChangedAction()
    {
        return [this](auto event) { onHomingChanged = event; };
    }

    auto MovingChangedAction()
    {
        return [this](auto event) { onMovingChanged = event; };
    }

    std::function<void()> onHomingChanged;
    std::function<void()> onMovingChanged;
};

struct MotorGlueEventsMock
{
    MOCK_METHOD(void, OnHomingChanged, ());
    MOCK_METHOD(void, OnMovingChanged, ());
};

struct MotorGlueComponentTest : testing::Test {
    MotorGlueComponentTest() :
        glue(locator
            .set(runtime)
            .set(pump)
            .set(static_cast<Hal::Motor&>(motor), motorId)
        )
    {
        glue.api.out.OnHomingChanged = [this]
        {
            events.OnHomingChanged();
        };
        glue.api.out.OnMovingChanged = [this]
        {
            events.OnMovingChanged();
        };

        dzn::check_bindings(glue);
    }

    ~MotorGlueComponentTest() override
    {
        pump.stop();
    }

    const std::string motorId = "abc";
    dzn::locator locator;
    dzn::runtime runtime;
    dzn::pump pump;
    testing::StrictMock<MotorMock> motor;
    testing::StrictMock<MotorGlueEventsMock> events;
    MotorGlueComponent glue;
};

struct MotorGlueComponentTest_Operational : MotorGlueComponentTest
{
    void SetUp() override
    {
        EXPECT_CALL(motor, HomingChanged(NotNull())).WillOnce(motor.HomingChangedAction());
        EXPECT_CALL(motor, MovingChanged(NotNull())).WillOnce(motor.MovingChangedAction());
        ASSERT_EQ(Result::Ok, dzn::shell(pump, glue.api.in.Initialize, motorId));

        ASSERT_TRUE(::testing::Mock::VerifyAndClear(&motor));
    }
};

}

TEST_F(MotorGlueComponentTest, InitializeSucceedsWhenItFindsMotor)
{
    EXPECT_CALL(motor, HomingChanged(NotNull()));
    EXPECT_CALL(motor, MovingChanged(NotNull()));
    ASSERT_EQ(Result::Ok, dzn::shell(pump, glue.api.in.Initialize, motorId));
}

TEST_F(MotorGlueComponentTest, InitializeFailsWhenItDoesNotFindMotor)
{
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.Initialize, "wrongId"));
}

TEST_F(MotorGlueComponentTest, Terminate)
{
    dzn::shell(pump, glue.api.in.Terminate);
}

TEST_F(MotorGlueComponentTest_Operational, Terminate)
{
    EXPECT_CALL(motor, HomingChanged(IsNull()));
    EXPECT_CALL(motor, MovingChanged(IsNull()));
    dzn::shell(pump, glue.api.in.Terminate);
}

TEST_F(MotorGlueComponentTest_Operational, HomeReturnsOk)
{
    EXPECT_CALL(motor, Home());
    ASSERT_EQ(Result::Ok, dzn::shell(pump, glue.api.in.Home));
}

TEST_F(MotorGlueComponentTest_Operational, HomeReturnsFail)
{
    EXPECT_CALL(motor, Home()).WillOnce(Throw(std::runtime_error("Home failed")));
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.Home));
}

TEST_F(MotorGlueComponentTest_Operational, IsHomingReturnsTrue)
{
    EXPECT_CALL(motor, IsHoming()).WillOnce(Return(true));
    ASSERT_EQ(Result::True, dzn::shell(pump, glue.api.in.IsHoming));
}

TEST_F(MotorGlueComponentTest_Operational, IsHomingReturnsFalse)
{
    EXPECT_CALL(motor, IsHoming()).WillOnce(Return(false));
    ASSERT_EQ(Result::False, dzn::shell(pump, glue.api.in.IsHoming));
}

TEST_F(MotorGlueComponentTest_Operational, IsHomingReturnsFail)
{
    EXPECT_CALL(motor, IsHoming()).WillOnce(Throw(std::runtime_error("IsHoming failed")));
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.IsHoming));
}

TEST_F(MotorGlueComponentTest_Operational, MoveReturnsOk)
{
    double value = 1234.0;
    EXPECT_CALL(motor, Move(value));
    ASSERT_EQ(Result::Ok, dzn::shell(pump, glue.api.in.Move, value));
}

TEST_F(MotorGlueComponentTest_Operational, MoveReturnsFail)
{
    double value = 5678.0;
    EXPECT_CALL(motor, Move(value)).WillOnce(Throw(std::runtime_error("Move failed")));
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.Move, value));
}

TEST_F(MotorGlueComponentTest_Operational, IsMovingReturnsTrue)
{
    EXPECT_CALL(motor, IsMoving()).WillOnce(Return(true));
    ASSERT_EQ(Result::True, dzn::shell(pump, glue.api.in.IsMoving));
}

TEST_F(MotorGlueComponentTest_Operational, IsMovingReturnsFalse)
{
    EXPECT_CALL(motor, IsMoving()).WillOnce(Return(false));
    ASSERT_EQ(Result::False, dzn::shell(pump, glue.api.in.IsMoving));
}

TEST_F(MotorGlueComponentTest_Operational, IsMovingReturnsFail)
{
    EXPECT_CALL(motor, IsMoving()).WillOnce(Throw(std::runtime_error("IsMoving failed")));
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.IsMoving));
}

TEST_F(MotorGlueComponentTest_Operational, StopReturnsOk)
{
    EXPECT_CALL(motor, Stop());
    ASSERT_EQ(Result::Ok, dzn::shell(pump, glue.api.in.Stop));
}

TEST_F(MotorGlueComponentTest_Operational, StopReturnsFail)
{
    EXPECT_CALL(motor, Stop()).WillOnce(Throw(std::runtime_error("Stop failed")));
    ASSERT_EQ(Result::Fail, dzn::shell(pump, glue.api.in.Stop));
}

TEST_F(MotorGlueComponentTest_Operational, OnHomingChangedEvent)
{
    EXPECT_CALL(events, OnHomingChanged());
    motor.onHomingChanged();
}

TEST_F(MotorGlueComponentTest_Operational, OnMovingChangedEvent)
{
    EXPECT_CALL(events, OnMovingChanged());
    motor.onMovingChanged();
}
